 /**
  * stream_client.c:
  * OpenCV video streaming client
  *
  * Author  Nash <me_at_nashruddin.com>
  *         gartim gartim_at_gmail.com
  * todo:
  * - add tcp connect interface to new port 8889, same arrowkeys to control steppers
  *
  * See the tutorial at
  * http://nashruddin.com/StrEAMinG_oPENcv_vIdEos_ovER_tHe_nEtWoRk
  */
  
#define VIDEO_PORT 8888
#define CNTL_PORT  8889

 
#include <sys/types.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include "cv.h"
#include "highgui.h"
  
IplImage* img;
int       is_data_ready = 0;
int       sock;
char*     server_ip;
  
 pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
  
 void* streamClient(void* arg);
 void  quit(char* msg, int retval);
  
 int main(int argc, char** argv)
 {
     pthread_t thread_c;
     int width, height, key;
  
     if (argc != 4) {
         quit("Usage: stream_client <server_ip> <width> <height>", 0);
     }
  
     /* get the parameters */
     server_ip   = argv[1];
     width       = atoi(argv[2]);
     height      = atoi(argv[3]);
  
     /* create image */
     img = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, 1);
     cvZero(img);
  
     /* run the streaming client as a separate thread */
     if (pthread_create(&thread_c, NULL, streamClient, NULL)) {
         quit("pthread_create failed.", 1);
     }
  
     fprintf(stdout, "Press 'q' to quit.\n\n");
     cvNamedWindow("stream_client", CV_WINDOW_AUTOSIZE);
  
     int fdnet = tcp_connect(argv[1]);
     printf("fdnet = %d\n", fdnet);

     while( 1 ) {
	switch ( key ) {
		case 'q':
                        printf("shutting down\n");
			break;
                case 'r':
                        printf("turning right\n");
			write(fdnet, &key, sizeof key);
                        break;
                case 'l':
                        printf("turning left\n");
			write(fdnet, &key, sizeof key);
                        break;
                case 'u':
                        printf("moving up\n");
			write(fdnet, &key, sizeof key);
                        break;
                case 'd':
                        printf("moving down\n");
			write(fdnet, &key, sizeof key);
                        break;
                case 'f':
                        printf("moving forward\n");
			write(fdnet, &key, sizeof key);
                        break;
                case 'b':
                        printf("moving back\n");
			write(fdnet, &key, sizeof key);
                        break;
                case 'e':
                        printf("moving forward left\n");
			write(fdnet, &key, sizeof key);
                        break;
                case 'i':
                        printf("moving forward right\n");
			write(fdnet, &key, sizeof key);
                        break;
                case 's':
                        printf("moving stopping all motors\n");
			write(fdnet, &key, sizeof key);
                        break;
                case 't':
                        printf("moving testing incr +100 to 2400\n");
			write(fdnet, &key, sizeof key);
                        break;
                case '?':
                        printf("unknown option: %c\n", key);
                        break;
                }
     	//read (fdnet, &key, sizeof key);
	if (key == 'q') break;

         /**
          * Display the received image, make it thread safe
          * by enclosing it using pthread_mutex_lock
          */
         pthread_mutex_lock(&mutex);
         if (is_data_ready) {
             cvShowImage("stream_client", img);
             is_data_ready = 0;
         }
         pthread_mutex_unlock(&mutex);
  
         key = cvWaitKey(10);
     }
  
     /* user has pressed 'q', terminate the streaming client */
     if (pthread_cancel(thread_c)) {
         quit("pthread_cancel failed.", 1);
     }
  
     /* free memory */
     cvDestroyWindow("stream_client");
     quit(NULL, 0);
 }
  
 /**
  * This is the streaming client, run as separate thread
  */
 void* streamClient(void* arg)
 {
     struct  sockaddr_in server;
  
     /* make this thread cancellable using pthread_cancel() */
     pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
     pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
  
     /* create socket */
     if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
         quit("socket() failed.", 1);
     }
  
     /* setup server parameters */
     memset(&server, 0, sizeof(server));
     server.sin_family = AF_INET;
     server.sin_addr.s_addr = inet_addr(server_ip);
     server.sin_port = htons(VIDEO_PORT);
  
     /* connect to server */
     if (connect(sock, (struct sockaddr*)&server, sizeof(server)) < 0) {
         quit("connect() failed.", 1);
     }
  
     int  imgsize = img->imageSize;
     char sockdata[imgsize];
     int  i, j, k, bytes;
  
     /* start receiving images */
     while(1)
     {
         /* get raw data */
         for (i = 0; i < imgsize; i += bytes) {
             if ((bytes = recv(sock, sockdata + i, imgsize - i, 0)) == -1) {
                 quit("recv failed", 1);
             }
         }
  
         /* convert the received data to OpenCV's IplImage format, thread safe */
         pthread_mutex_lock(&mutex);
  
         for (i = 0, k = 0; i < img->height; i++) {
             for (j = 0; j < img->width; j++) {
                 ((uchar*)(img->imageData + i * img->widthStep))[j] = sockdata[k++];
             }
         }
  
         is_data_ready = 1;
         pthread_mutex_unlock(&mutex);
  
         /* have we terminated yet? */
         pthread_testcancel();
  
         /* no, take a rest for a while */
         usleep(1000);
     }
 }
  
 /**
  * This function provides a way to exit nicely from the system
  */
 void quit(char* msg, int retval)
 {
     if (retval == 0) {
         fprintf(stdout, (msg == NULL ? "" : msg));
         fprintf(stdout, "\n");
     } else {
         fprintf(stderr, (msg == NULL ? "" : msg));
         fprintf(stderr, "\n");
     }
  
     if (sock) close(sock);
     if (img) cvReleaseImage(&img);
  
     pthread_mutex_destroy(&mutex);
  
     exit(retval);
 }
  
int tcp_connect(char *host)
{ 
    struct hostent *host_info;
    struct servent *serv_info;
    struct sockaddr_in  server;
    int sock;
 
    /* Create a socket */
    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        quit("socket() failed.", 1);
    }
    /* setup server parameters */
    memset(&server, 0, sizeof(server));
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = inet_addr(server_ip);
    server.sin_port = htons(CNTL_PORT);
  
    /* Now make the connection */
    if (connect(sock, (struct sockaddr*)&server, sizeof(server)) < 0) {
        quit("connect() failed.", 1);
    }
    return sock;
}
